//HGCI subsystem

import java.util.Map;
import java.util.Iterator;
import SimpleOpenNI.*;

SimpleOpenNI context;
Map<Integer,ArrayList<PVector>>  handPathList = new HashMap<Integer,ArrayList<PVector>>();
PImage handIcon;
int handVecListSize = 3;
float zoomF =0.5f;
float rotX = radians(180);  
float rotY = radians(0);
boolean twoHandsMode = false;
PVector dir = null, midPoint = null;
float mag = 0;
float scale0=1, scale1=1, scale2=1;
float translate0=0, translate1=0, translate2=0;
float returnRotX=0, returnRotY=0, returnRotZ=0;

float default_distance = 600;//how far are you from kinect
float threshold1 = 200; //move too past
float threshold2 = 5; //make it stable

void setupHGCI(){
  context = new SimpleOpenNI(this);
  if(context.isInit() == false)
  {
     println("Can't init SimpleOpenNI, maybe the camera is not connected!"); 
     exit();
     return;  
  }   
  
  context.enableDepth();
  context.setMirror(true);
  
  context.enableHand();
  context.startGesture(SimpleOpenNI.GESTURE_WAVE); //GESTURE_WAVE
  context.startGesture(SimpleOpenNI.GESTURE_CLICK);
  
  handIcon = loadImage("handIcon.png");
  
  //camera(width, -height, 700, 0, 0, 0, 0, 1, 0);
  //camera(width, height, height / tan(PI*30.0 / 180.0), width, height, 0, 0, 1, 0);
  
  //perspective(radians(45),
  //            float(width)/float(height),
  //            10.0f,150000.0f);
}

void HGCI(){
  //clear(); 
  context.update();
  
  // set the scene pos
//  translate(width/2, height/2, 0);
//  rotateX(rotX);
//  rotateY(rotY);
//  scale(zoomF);
   
   if(handPathList.size() > 0){
     Iterator it = handPathList.entrySet().iterator();
     //int handsCounter = 0;
     ArrayList<PVector> handsCPos = new ArrayList<PVector>();
     while(it.hasNext()){
       Map.Entry mapEntry = (Map.Entry)it.next();
       int handId =  (Integer)mapEntry.getKey();
       ArrayList<PVector> vecList = (ArrayList<PVector>)mapEntry.getValue();
       PVector p, previousP, previousP2;
       p = vecList.get(0); 
       if(vecList.size() > 2){
         previousP = vecList.get(1);
         previousP2 = vecList.get(2);
         PVector newPoint = filterData(handId, p, previousP, previousP2);
         if(newPoint != null){
           handsCPos.add(newPoint);
         }
       }else{
         handsCPos.add(p);
       }
     }  
     
     if(handsCPos.size() == 2){
       if(!twoHandsMode){
          //set standard
          mag = v_dist(handsCPos.get(0), handsCPos.get(1));
          dir = v_sub(handsCPos.get(0), handsCPos.get(1));
          midPoint = v_mul(v_add(handsCPos.get(0), handsCPos.get(1)), 0.5);
          twoHandsMode = true;
          println("enter two hands mode.");
          println("dir: " + dir.x + "  " +dir.y +"  "+ dir.z);
          println("mag: "+ mag);
          println("midPoint: " + midPoint.x + "  " +midPoint.y +"  "+ midPoint.z);
       }else{
         //two hands mode
         //calculate
         float Cmag = v_dist(handsCPos.get(0), handsCPos.get(1));
         PVector CmidPoint = v_mul(v_add(handsCPos.get(0), handsCPos.get(1)), 0.5);
         PVector Cdir = v_sub(handsCPos.get(0), handsCPos.get(1));
         scale0 = (float)Cmag/mag;
         scale1 = (float)Cmag/mag;
         scale2 = (float)Cmag/mag;
       
         translate0 = CmidPoint.x - midPoint.x;
         translate1 = CmidPoint.y - midPoint.y;
         translate2 = CmidPoint.z - midPoint.z;
         
         if(Cdir.z != 0){
           if(Cdir.x > 0){
             returnRotY = asin(Cdir.z/Cmag);
           }else{
             returnRotY = -asin(Cdir.z/Cmag);
           }
         }
         if(Cdir.y != 0){
           if(Cdir.x > 0){
             returnRotZ = asin(Cdir.y/Cmag);
           }else{
             returnRotZ = -asin(Cdir.y/Cmag);
           }
         }
       }
     }else{
       twoHandsMode = false;
     }
   }  
}

void showHands(){
  pushMatrix();
  translate(0,0,default_distance);
  rotateX(rotX);
      //rotateY(rotY);
  scale(zoomF);
  //draw 2 hands
  imageMode(CENTER);
  stroke(255);
  strokeWeight(2);
  if(handPathList.size() > 0){
     Iterator it = handPathList.entrySet().iterator();
     while(it.hasNext()){
       Map.Entry mapEntry = (Map.Entry)it.next();
       int handId =  (Integer)mapEntry.getKey();
       ArrayList<PVector> vecList = (ArrayList<PVector>)mapEntry.getValue();
       PVector p;
       p = vecList.get(0);
       pushMatrix();
         translate(p.x, p.y, p.z);
         rotateX(rotX);
         image(handIcon, 0, 0, 50, 50);
       popMatrix(); 
     }  
  }
  popMatrix();
  
  
}

void beginHGCI(){
  pushMatrix();
    perspective(radians(45),
                float(width)/float(height),
                10.0f,150000.0f);
    
    pushMatrix();
      //translate(width/2, height/2, 0);
      rotateX(rotX);
        //rotateY(rotY);
      scale(zoomF);
      
      scale(scale0, scale1, scale2);
      rotateY(returnRotY);
      rotateZ(returnRotZ);
      rotateX(rotX);
      translate(translate0, -translate1, translate2);
}


void endHGCI(){
    popMatrix();
  popMatrix();
}


void HGCIwithGUI(){
  float fov = PI/3.0;
  perspective(fov, float(width)/float(height), 10.0f, 150000.0f);
  
}

PVector filterData(int handID, PVector p, PVector previousP, PVector previousP2){
  PVector average = v_mul(v_add(p, previousP),0.5);
  PVector speedP = v_sub(previousP, previousP2);
  PVector speedC = v_sub(p, previousP);
  
  println("test data: " + v_len(speedC));
  if(v_len(speedC) > threshold1){
    println("over threshold1: too fast -- " + v_len(speedC));
    return null;
  }
  
  if(v_len(speedC) < threshold2 && v_len(speedC) < v_len(speedP)){
    println("over threshold2: stable " + v_len(speedC));
    return new PVector(previousP.x, previousP.y, previousP.z);
  }
  
  return new PVector(p.x, p.y, p.z);
}

void onNewHand(SimpleOpenNI curContext,int handId,PVector pos){
  println("onNewHand - handId: " + handId + ", pos: " + pos);
 
  ArrayList<PVector> vecList = new ArrayList<PVector>();
  vecList.add(pos);
  
  handPathList.put(handId,vecList);
}

void onTrackedHand(SimpleOpenNI curContext,int handId,PVector pos){
  //println("onTrackedHand - handId: " + handId + ", pos: " + pos );
  
  ArrayList<PVector> vecList = handPathList.get(handId);
  if(vecList != null)
  {
    vecList.add(0,pos);
    if(vecList.size() > handVecListSize)
      // remove the last point 
      vecList.remove(vecList.size()-1); 
  }
  
}


void onLostHand(SimpleOpenNI curContext,int handId){
  println("onLostHand - handId: " + handId);
  handPathList.remove(handId);
 
}


void onCompletedGesture(SimpleOpenNI curContext, int gestureType, PVector pos)
{
  println("onCompletedGesture - gestureType: " + gestureType + ", pos: " + pos);
  
  int handId = context.startTrackingHand(pos);
  
  println("hand stracked: " + handId);
  
}

